
/*****************************************************************************/
/*                                                                           */
/*                	National Instruments GPIB-VXI Project                    */
/*						DownLoable Code Instruments							 */
/*                                                                           */
/*                     (C) 1989  National Instruments                        */
/*                                                                           */
/*****************************************************************************/
#include "stdio.h"
#include "DataSize.h"
#include "ResManIn.h"
#include "CommonBu.h"
#include "DciStruc.h"
#include "DciFuncs.h"

/*
 *	This file contains the I/O drivers to communicate with the Colorado Data
 * Systems 852 Adapter Module.  The functions needed are init, reset, read, 
 * write and local command set modes.  These functions provide all of the high
 * level functionality needed to make the 852 act like a VXI Message Based
 * device.
 */


	/* the following are offset addresses and bit postions for 852 registers */
#define CDS852_ID_PROM			0x000
#define CDS852_ACTIVE			0x100
#define CDS852_PASSIVE			0x101
#define CDS852_INT_ACK			0x200

#define CDS852_CONTROL			0x300
#define CDS852_CONT_DISABLE		(1<<2)
#define CDS852_CONT_NO_RESET	(1<<1)

#define CDS852_STATUS			0x300
#define CDS852_STAT_READY		(1<<4)


#define CDS852_END				(1<<8)		/* End of Message (byte | 0x100) */
#define CDS852_MAX_READ			0x200		/* best when mulitple of 0x100 (smallest alloc size)*/


extern int8 MyLA;
extern int8 SecondaryAddress;

uint16 *base852;				/* holds A24 base address of the 852 module */
uint32 BinSize;					/* read size for binary reads */
int8 BinMode;					/* 1 = binary read mode, 0 = ASCII read mode */
uint16 CDS852_EOS;				/* holds ASCII End of string character */
uint16 CDS852_EOT;				/* 1 = end on bit 8 of read, 0 = do not*/
int8 DebugMode;					/* 1 = Debug mode on, 0 = off */

int8 InIO, CLEAR;				/* flags to abort I/O in progress on a Clear */




/*
 *    This Init routine is used to set up default variable settings and to
 * reset the 852 module.
 */
Init852()
{
		/* default 852 address is DCI logical address = upper 8 bits
		 * of A24 address
		 */
	base852 = (int8 *)(MyLA << 16);
	BinMode = 0;
	BinSize = CDS852_MAX_READ;
	DebugMode = DCIheader.Debug? 1 : 0;
	CDS852_EOS = 0x0a;
	CDS852_EOT = 1;

	Reset852();
}



/*
 *   The Reset routine is used to reset the 852.  This is done by enabling
 * the 852, putting it in reset state for a period of time, and then
 * releasing it from reset state.  If a Bus Error occurs accessing the 
 * control/status register then it is assumed that no 852 module resides
 * at that A24 address.  The address is changed to 'invalid address'.
 */
Reset852()
{
	uint16 *control;
	uint16 tester;


	if (base852 == (uint16 *)0xffffffff)
	{ 	if (DebugMode)
		{	printf("Unable to reset 852, A24 address is invalid\r\n");
		}
		return 0;
	}
	
		/* calculate the base address of the control/status register */
	control = base852 + CDS852_CONTROL;

	if (!(SafeReadWord(control,&tester)))
	{
		if (DebugMode)
		{	printf("CDS 852 CI got Bus Error reading from 852\r\n");
			printf("Assuming invalid 852 location (base was A24 address 0x%x) \r\n",base852);
		}

		base852 = (uint16 *)0xffffffff;
		return;
	}

		/* put 852 enabled and reset, wait 1/25 second, unreset it and disable it */
	*control = 0x0000;
	pause_p(4);	
	*control = CDS852_CONT_NO_RESET | CDS852_CONT_DISABLE;

	if (DebugMode)
	{	printf("\r\nCDS 852 CI reset 852 at A24 address 0x%x\r\n",(uint32)base852);
	}

	if (InIO)
	{	CLEAR = 1;
	}
}




/*
 *	The following is the read routine for the 852 module.  It will read up to
 * MaxRead or BinSize amount of bytes depending on the END modes set.  End Of
 * String (EOS) modes apply on to ASCII but END bit mode applies to both Binary
 * and ASCII mode.  The buffer returned is a CommonBuf so that it can be easily
 * passed out the Word Serial to the Commander or GPIB.
 */
CommonBuf *Read852()
{
	uint16 *base, *control, *stat, *active, *passive;
	uint16 tempword, eos;
	CommonBuf *ReadBuf;
	uint32 size, i;


	if (base852 == (uint16 *)0xffffffff)
	{	if (DebugMode)
		{	printf("CI %d NOT reading from 852 (Invalid 852 address) \r\n",MyLA);
		}
		return 0;
	}

	size = BinMode? BinSize : CDS852_MAX_READ;

	eos = BinMode? 0xffff : CDS852_EOS;		/* since 8 bit compare 0xffff fails */

		/* Create CommonBuf large enough to hold max read in data */
	ReadBuf = CreateBuf( ((size + 0xff) >> 8), 0);

		/* Pre-compute addresses for speed */
	stat 	= base852 + CDS852_STATUS;
	active 	= base852 + CDS852_ACTIVE;
	passive	= base852 + CDS852_PASSIVE;

		/* enable the 852 module */
	control = base852 + CDS852_CONTROL;
	*control = CDS852_CONT_NO_RESET;

	if (DebugMode)
	{	printf("CI %d: read up to 0x%4x (%4d) bytes (%s) from CDS 852.....",
			MyLA, size, size, BinMode? "Bin":"ASCII");
	}

	InIO = 1;

	for (i = 0; (i < size); i++)
	{
		tempword = *active;			/* initiate read */
		for (;;)
		{
				/* A "forever" loop with multiple tests is MUCH faster than any
				 * other type of loop structure for polled I/O.  This is because
				 * the majority of time is spent testing the Ready condition, not
				 * executing loop overhead code.
				 */
			if (*stat & CDS852_STAT_READY)  break;
			if (*stat & CDS852_STAT_READY)  break;
			if (*stat & CDS852_STAT_READY)  break;
			if (*stat & CDS852_STAT_READY)  break;
			if (*stat & CDS852_STAT_READY)  break;
			if (*stat & CDS852_STAT_READY)  break;
			if (*stat & CDS852_STAT_READY)  break;
			if (*stat & CDS852_STAT_READY)  break;
			if (*stat & CDS852_STAT_READY)  break;
			if (*stat & CDS852_STAT_READY)  break;

			if (CLEAR)
			{	if (DebugMode)
				{	printf("CLEAR --> Read from 852 aborted. \r\n");
				}
					/* disable the 852 module */
				*control = CDS852_CONT_NO_RESET | CDS852_CONT_DISABLE;

				tempword = free_seg(ReadBuf);
				CLEAR = 0;
				InIO = 0;
				return 0L;
			}

				/* give up processor temporarily so other processes may run */
			pause_p(0);
		}

		tempword = *passive;					/* read WORD and */
		ReadBuf->data[i] = (int8)tempword;		/* store lower byte in buffer */


		if (CDS852_EOS == 0xffff)
		{		/* end on END bit only */
			if (CDS852_EOT && (tempword & CDS852_END))
			{	i++;
				break;
			}
			else
			{
				/* do not end, not Terminal count or not END (when enabled) */
			}
		}
		else
		{		/* end on END and EOS */
			if (CDS852_EOT)
			{	if ((tempword & CDS852_END) && ((tempword & 0xff) == CDS852_EOS))
				{	i++;
					break;
				}
				else
				{
					/* END but not EOS */
				}
			}
			else
			{		/* no END enabled and got EOS */
				if ((tempword & 0xff) == CDS852_EOS)
				{	i++;
					break;
				}
				else
				{
					/* no END enabled and not EOS */
				}
			}
		}
	}
	ReadBuf->EndFlag = 1;
	ReadBuf->data[i] = '\0';
	ReadBuf->byteCount = i;

	if (DebugMode)
	{	printf("Done (%d bytes)\r\n",ReadBuf->byteCount);
		if (ReadBuf->byteCount < 512)
		{	printf("%s\r\n",ReadBuf->data);
		}
	}

		/* disable the 852 module */
	*control = CDS852_CONT_NO_RESET | CDS852_CONT_DISABLE;

	InIO = 0;
	CLEAR = 0;

	return ReadBuf;
}




/*
 *	The following is the write routine for the 852 module.  It takes in a CommonBuf
 * and sends it to the 852 a byte at a time.  It the EndFlag is set in the CommonBuf
 * it sends END (bit 8) with the last byte of the transfer.
 */
Write852(OutBuf)
	CommonBuf *OutBuf;
{
	uint16 *base, *control, *stat, *active, *passive;
	uint16 tempword;
	uint32 count, i;
	int8 sendend;


	if (base852 == (uint16 *)0xffffffff)
	{	if (DebugMode)
		{	printf("CI %d throwing away 0x%4x (%4d) byte buffer In (Invalid 852 address) \r\n",
				MyLA,OutBuf->byteCount,OutBuf->byteCount);
		}
		return 0;
	}

		/* Pre-compute addresses for speed */
	stat = base852 + CDS852_STATUS;
	active = base852 + CDS852_ACTIVE;
	passive= base852 + CDS852_PASSIVE;

		/* enable the 852 module */
	control = base852 + CDS852_CONTROL;
	*control = CDS852_CONT_NO_RESET;

	count = OutBuf->byteCount - 1;
	sendend = OutBuf->EndFlag;

	if (DebugMode)
	{	printf("CI %d: trying to write 0x%4x (%4d) bytes to 852",MyLA,count+1,count+1);
		if (OutBuf->byteCount < 512)
		{	printf(":\r\n%s\r\n",OutBuf->data);
		}
		else
		{	printf(".....");
		}
	}

	InIO = 1;

	for (i = 0; (i <= count); i++)
	{
		tempword = (uint16)OutBuf->data[i];

		if (sendend && (i >= count))
			tempword |= CDS852_END;

		*active = tempword;			/* write out word */

		for (;;)
		{
				/* A "forever" loop with multiple tests is MUCH faster than any
				 * other type of loop structure for polled I/O.  This is because
				 * the majority of time is spent testing the Ready condition, not
				 * executing loop overhead code.
				 */
			if (*stat & CDS852_STAT_READY)  break;
			if (*stat & CDS852_STAT_READY)  break;
			if (*stat & CDS852_STAT_READY)  break;
			if (*stat & CDS852_STAT_READY)  break;
			if (*stat & CDS852_STAT_READY)  break;
			if (*stat & CDS852_STAT_READY)  break;
			if (*stat & CDS852_STAT_READY)  break;
			if (*stat & CDS852_STAT_READY)  break;
			if (*stat & CDS852_STAT_READY)  break;
			if (*stat & CDS852_STAT_READY)  break;

			if (CLEAR)
			{	if (DebugMode)
				{	printf("CLEAR --> Write to 852 aborted. \r\n");
				}
					/* disable the 852 module */
				*control = CDS852_CONT_NO_RESET | CDS852_CONT_DISABLE;

				CLEAR = 0;
				InIO = 0;
				return 0L;
			}

				/* give up processor temporarily so other processes may run */
			pause_p(0);
		}
		/* byte finished being written out */
	}

		/* disable the 852 module */
	*control = CDS852_CONT_NO_RESET | CDS852_CONT_DISABLE;

	if (DebugMode)
	{	printf("Done \r\n");
	}

	InIO = 0;
	CLEAR = 0;
}




/*
 *	Local commands are commands that will not be routed to the 852 module.
 * They will instead be interpretted by the DCI itself.  A buffer beginning
 * with "!!" means local command.  This was done so as not to interfere with
 * any 852 module's command set.  A single letter following the "!!" is the
 * command to execute.  The commands may have parameters.  The commands are 
 * as follows:
 *		A					set ASCII read mode (read till END char)
 *		B					set BINARY read mode (read max number)
 *		S <decimal number>	set BINARY read SIZE
 *		L <hex number>		change address of 852 module (upper 8 bits A24)
 *		Q,q					query information about CI settings
 *		D 					enable  debug printing 
 *		d 					disable debug printing 
 *		E,e <hex number>	set character to end read on (EOS character)
 *		T					enable ending read on END bit (8th bit) 
 *		t					disable ending read on END bit (8th bit)
 */		

ExecuteLocalCommand(InBuf)
	CommonBuf *InBuf;
{
	uint32 size;
	uint32 mode;

	switch (InBuf->data[2])
	{
		case 'd':	
			if (DebugMode)
			{	printf("\r\nDebug Mode now turned off.\r\n");
			}
			DebugMode = 0;
			break;

		case 'D':
			DebugMode = 1;
			printf("\r\nDEBUG MODE ON..... \r\n");

		case 'q':	case 'Q':
			printf("\n\r");
			printf("CDS 852 Code Instrument information: \r\n");
			printf("   Logical Address      0x%x \r\n",MyLA);
			printf("   Seconary Address     %d  \r\n",SecondaryAddress);
			printf("   A24 base address     0x%x %s\r\n",
				(uint32)base852, ((uint32)base852 == 0xffffffff)?"(Not Valid)":"");
			printf("   Data Read Mode       %s \r\n",
				BinMode? "Binary" : "ASCII");
			printf("   Binary Read Size     0x%x \r\n", BinSize);
			printf("   End of String char   0x%x %s\r\n", 
				CDS852_EOS, (CDS852_EOS == 0xffff)?"(none)":""); 
			printf("   End Read on Bit 8?   %s \r\n", CDS852_EOT? "Yes":"No");
			printf("\n\r");
			break;

		case 'l':	case 'L':
			sscanf(&(InBuf->data[3]),"%x",&size);
			if (size >= 0x20 && size <= 0xDf)
			{	base852 = (int8 *)(size << 16);
				Reset852();
			}
			else
			{	printf("\r\n");
				printf("Address must be 0x20-0xDF (0x200000-0xDF0000) \r\n");
				printf("GPIB-VXI cannot see A24 addresses outside this range \r\n");
			}

			break;

		case 'b': 	case 'B':
			BinMode = 1;
			if (DebugMode)
			{	printf("Binary Read Mode Enabled for CI %d\r\n",MyLA);
			}
			break;

		case 'a': 	case 'A':
			BinMode = 0;
			if (DebugMode)
			{	printf("ASCII Read Mode Enabled for CI %d\r\n",MyLA);
			}
			break;

		case 's': 	case 'S':
			sscanf(&(InBuf->data[3]),"%d",&size);
			if (size != 0 && size < 0x20000)
				BinSize = size;
			if (DebugMode)
			{	printf("Binary Read Size = 0x%x \r\n", BinSize);
			}
			break;

		case 'e':	case 'E':
			sscanf(&(InBuf->data[3]),"%x",&size);
			if (size < 0x100)
				CDS852_EOS = size;
			else
				CDS852_EOS = 0xffff;
			if (DebugMode)
			{	printf("852 End of String Character is 0x%x %s\r\n",
					CDS852_EOS, (CDS852_EOS == 0xffff)?"(None)":"");
			}
			break;

		case 't':	
			CDS852_EOT = 0;
			if (DebugMode)
			{	printf("852 End on bit Bit 8 of response OFF \r\n");
			}
			break;

		case 'T':
			CDS852_EOT = 1;
			if (DebugMode)
			{	printf("852 End on bit Bit 8 of response ON \r\n");
			}
			break;

		default:
			printf("CDS 852 CI %d: ILLEGAL LOCAL COMMAND. \r\n",
				MyLA);
	}
}


