
/*****************************************************************************/
/*                                                                           */
/*                	National Instruments GPIB-VXI Project                    */
/*						DownLoable Code Instruments							 */
/*                                                                           */
/*                     (C) 1992  National Instruments                        */
/*                                                                           */
/*****************************************************************************/
/*	The following file contains all of the functions needed by the Asynch
 * process.
 *	The purpose of the Asynch process is to handle all asynchronous events.
 * These events include Word Serial commands and responses, Event and Response
 * signals (and VXI interrupts in the form of signals), handling a servant
 * that has FAILED (has been SYSFAIL Inhibited), Control register simulation,
 * and recording the fact that a Word Serial Data buffer came in.
 *
 *	NO EXTENED PROCESSING should be done in the Asynch process.  The only way
 * that the Asynch process can act in an asychronous matter is if it is waiting
 * for asynchronous messages to come in MOST of the time.  If the asychronous
 * message causes a need for extended processing, it should signal the Worker
 * process or another spawed off process to do the work for it.
 */
/*****************************************************************************/

#include "stdio.h"
#include "DataSize.h"
#include "ResManIn.h"
#include "S_EVENT.h"
#include "DCI_DEBU.h"
#include "CommonBu.h"
#include "InclWS.h"
#include "signals.h"
#include "DciStruc.h"
#include "DciFuncs.h"


extern int8 MyLA, Cmdr;				/* direct global used for speed */
extern uint16 ParserId;				/* direct global used for speed */
extern uint16 *VXIregs;
extern la_entry *rmtable;			/* global pointer to Resource Manager table */
extern int8 GPIB_VXI_LA;

uint32 AsynchXid, WorkerPid;		/* used for quick reference to ID values */
uint16 OurServantArea = 5;			/* servant area value...only if applicable */
int8 BNOtype;						/* keeps track of when/what type of Begin Normal Operation */
int8 NormalOperation;

	/* bit-wise event enables through Control Event command */
uint32 CIEventEnables[4];     /* [0]=0-31, [1]=32-63, [2]=64-95, [3]=96-127 */

int32 WS_DATA_IN_NUM;				/* Holds current number of WS Buffer In messages pending */
int8 GetOutput;						/* flag for worker to get output from the device */



void AsynchProcess()
{
	struct { uint32 u,h, type, p1, p2, p3} m;		/* pSOS message struct */


	start("0.0","0.0","0.0");					/* allow Serial port I/O */

	AsynchXid  = DCIheader.AsynchXid;
	WorkerPid = DCIheader.WorkerPid;

		/* init: no data in and do not get any data yet */
	WS_DATA_IN_NUM = 0L;
	GetOutput = 0;

	for(;;)
	{
		req_x(AsynchXid,&m,0,0);			/* Wait for asynchronous action to occur */
		switch (m.type)
		{
			case DCI_EVENT_Signal:
				HandleSignal((uint16)m.p1);
#ifdef RORA_INTERRUPTER
				if (m.p2 != 0)			/* p2 is the handler number */
					ReenableHandler(m.p2);
#endif
				break;

			case DCI_EVENT_WS_Command:
				HandleWordSerialCommand((uint16)m.p1);
				break;

			case DCI_EVENT_WS_Data:
					/* need to use number since signal_v()'s are not cumlative */
				WS_DATA_IN_NUM++;
				signal_v(WorkerPid, S_EVENT_DCI_IO);
				break;

			case DCI_EVENT_Control_In:
				printf("DCI EVENT: Control Register In--0x%x.\r\n",m.p1);
				break;

			case DCI_EVENT_WSphys_Command:
#ifdef NEED_PHYSICAL_WORD_SERIAL
				HandleWordSerialPhysCommand((uint16)m.p1);
#endif
				break;

			case DCI_EVENT_DOR_Request:
				GetOutput = 1;
				signal_v(WorkerPid, S_EVENT_DCI_IO);
				break;

			case DCI_EVENT_ServantFailed:
				printf("DCI EVENT: Servant at Logical Address %d FAILED.\r\n",(int8)m.p1);
					/* 
					 *	Any routines that need to be run in order to abort operation
					 * with the servant device and/or perform tests on the servant
					 * need to be added here.
					 */
				break;

			default:
				printf("DCI EVENT: Unknown Event received--0x%x.\r\n",m.type);
				break;
		}
	}
}




/*
 *	'HandleSignal' processes incoming VXI Event/Reponse signals.
 * Remember: a servant's signal/interrupt sending is controlled the Word
 * Serial commands: Asynchronous Mode Control, Reponse Enable (Control Response), 
 * and Control Event.
 *	If a DCI wishes its servants to be able to send signals to it, it should
 * send the appropriate Word Serial Commands to it at the beginning of the
 * Worker process or device setup time.  If the servant is a register based
 * device, then there should be some device dependent command or setup in order
 * to set the signal passing types and mode.
 */

HandleSignal(signal)
	uint16 signal;
{

	uint16 com,la;

#ifdef DEBUG_SIGNAL
	printf("DCI %d: Signal In: ",MyLA);
#endif

	com = signal & 0xff00;			/* signal type */
	la  = signal & 0x00ff;			/* logical address of sender (as per VXI specification) */

	/*
	 * NOTE: 
	 *	The following is formatted for handling a Message based device signaler/interrupter.
	 * A Register based interrupter will have the 'la' equal to its Logical Address also
	 * but the upper 8 bits 'com' are device dependent and should be treated as such.
	 */


	if (!(com & 0x8000))			/* Reponse Signal? */
	{
#ifdef DEBUG_SIGNAL
printf("Response signal 0x%x from LA %d \r\n", signal, la);
#endif
		if (com & Response_DOR)
		{
			/* A servant is now Data Out Ready */
		}


		if (com & Response_DIR)
		{
			/* A servant is now Data In Ready */
		}

			/***********************************************/
			/***** Handle other responses signals here *****/
			/***********************************************/

	}
	else                        /* Event Signals. */
	{   switch (com)
		{
		case SIG_REQT:
			/* Servant with logical address "la" sent you an REQT.
			 * if this means that you now need service, update your
			 * status byte and make a call to this function:
			 * 		SignalOut(Cmdr, SIG_REQT | MyLA);
			 * Otherwise, handle the servant's need for service
			 * locally in the CI.  If this causes a need for extended
			 * processing, signal the Worker (or another process) to
			 * handle that processing.
			 */	
#ifdef DEBUG_SIGNAL
printf("REQT from LA %d \r\n", la);
#endif
		break;


		case SIG_REQF:
			/* Servant with logical address "la" sent you an REQF.
			 * if this means that you now no longer need service, update 
			 * status byte and make a call to this function:
			 * 		SignalOut(Cmdr, SIG_REQF | MyLA);
			 * Otherwise, handle the servant's need for service
			 * locally in the DCI.  If this causes a need for extended
			 * processing, signal the Worker (or another process) to
			 * handle that processing.
			 */	
#ifdef DEBUG_SIGNAL
printf("REQF from LA %d \r\n", la);
#endif
        break;


		case SIG_UNREC_CMD:
				/* Do an Unrecognized Command Abort of WS command transfer
				 * (internally....done only if needed and for 1.2's only)
				 *	National Instruments will handle this signal for the CI
				 * if it makes the following call:
				 *		d_ctrl(DEV_WS,la,WS_CTRL_UnRecCom,0);
				 * If the DCI wishes to record the fact that a particular
				 * command was Unrecognized, the original caller of the
				 * Word Serial Send Command should do that (not the Asynch process).
				 */
#ifdef DEBUG_SIGNAL
printf("Unrecognized Command from LA %d \r\n", la);
#endif
			d_ctrl(DEV_WS,la,WS_CTRL_UnRecCom,0);
        break;


		case SIG_NO_CAUSE:
			/* if DCI has any VXI/VME Interrupter that is D08
			 *	only, this is the signal it will recieve when it pulls
			 *	its interrupt line.  "la" contains the D08 latched value.
			 *  "la" must be equal to a servant of the CI.
			 */
		break;


			/***********************************/
			/*  Handle any User defined signals here:
			 *
			 * 	According to the VXI Specification all signals of the form:
			 *
			 * bit position:	1 1 1 1 1 1 0 0 0 0 0 0 0 0 0 0 
			 *					5 4 3 2 1 0 9 8 7 6 5 4 3 2 1 0
			 *				----------------------------------------
			 *	value:			1 0 x x x x x x (logical address)
			 * 						^
			 * are USER definable.  National Instruments goes on to split
			 * this up according to bit 13.  If bit 13 is a "1", it is reserved
			 * for use by National Instruments only.  if bit 13 is a "0", it
			 * may be used for any purpose by a DCI and other message based
			 * or register based devices.  This leaves 32 signals for NI and
			 * 32 signals for DCI's.
			 */
			/***********************************/

 		}
	}
}




/*
 * 	'HandleWordSerialCommand' handles both the acceptance (and possibly
 * the response) for all of the Word Serial Commands.
 *	A CI may decide which subset of the VXI Message Based functionality 
 * it wishes to implement by the response (and actions) to these commands.
 * A CI must implement all of the commands required for minimal message
 * based devices according to the VXI specification.  It is HIGHLY recommended
 * that a CI implement all of the functionality allowed to a Message
 * Based device by the VXI specification.  The GPIB Secondary Address Model
 * will get much better throughput and the GPIB-VXI runtime system will be
 * more efficient if commands like Read STB, Response Enable, Asynchronous
 * Mode Control, etc. are implemented.  ALSO, there is no reason not to implement
 * these commands since most of the code (if not all) is supplied by National
 * Instruments in the following function already.
 *
 * NOTE: Variables whose contents are updated in these functions which are contained
 * in the CI Header or Resource Manager table MUST STAY AS IS.  These values 
 * are accessed directly from inside National Instruments onboard firmware 
 * (Word Serial driver, etc.).
 */

HandleWordSerialCommand(com)
	uint16 com;
{
	uint16 data, OurAsynMode, respmask, Our_RespEna;
	WSresp WSr;
	uint32 eventnum, eventind;


#ifdef DEBUG_WS_COMMAND
	printf("DCI %d -> WS Command: ",MyLA);
#endif

	data = com & 0x00ff;
	switch (com & 0xff00)
	{
		case WS_READ_STB & 0xff00:
#ifdef DEBUG_WS_COMMAND
			printf("Read Status Byte  ");
#endif
			/*
			 * This command returns the device status (could be a Serial Poll
			 * on GPIB).  Changing your status is as simple as changing the
			 * variable 'DCIheader.StatusByte'.  DO NOT CHANGE THIS FACT.  This 
			 * variable must be used to report status in (it may be internally
			 * accessed for quicker speed).
			 */
			WSr.Response = DCIheader.StatusByte | 0xff00;
			DCIheader.StatusByte = 0;			/* remove RSV and error */
			d_ctrl(DEV_WSsl,MyLA,WSsl_CTRL_Response,&WSr);
   
#ifdef DEBUG_WS_COMMAND
			printf("0x%x  (ret = 0x%x)\n\r", WSr.Response, WSr.RetVal);
#endif
			break;



		case WS_CLEAR & 0xff00:	
#ifdef DEBUG_WS_COMMAND
			printf("Device Clear.\n\r");
#endif
			DCIwsClear(MyLA);			/* Clear Own Word Serial I/O */
			WS_DATA_IN_NUM = 0;			/* Clear any pending WS buffer ops */
			ClearParser(ParserId);		/* Clear any pending commands/responses on parser */
			ClearServants();			/* Abort WS I/O with Servants and propagate CLEAR */

				/**** Handle any other Local CLEAR routines needed HERE ****/
			DCIheader.StatusByte = 0;			/* remove RSV and error */

			break;



		case WS_TRIGGER & 0xff00:	
#ifdef DEBUG_WS_COMMAND
			printf("Device Trigger.\n\r");
#endif
				/* Handle CI Word Serial Trigger Here */
			break;



		case WS_BEGIN_NORM & 0xff00:	
#ifdef DEBUG_WS_COMMAND
			printf("Begin Normal Operation. \n\r");
#endif

			if (NormalOperation)
			{		/* Set READY bit (PASSED bit should already be set) */
				VXIregs[StatusReg] |= Status_PASSED | Status_READY; 

				WSr.Response = 0x7ffe;
				d_ctrl(DEV_WSsl,MyLA,WSsl_CTRL_Response,&WSr);
			}
			else
			{	WSr.Response = PropagateBNO();
				BNOtype = 1;
				NormalOperation = 1;

					/* Set READY bit (PASSED bit should already be set) */
				VXIregs[StatusReg] |= Status_PASSED | Status_READY; 

				d_ctrl(DEV_WSsl,MyLA,WSsl_CTRL_Response,&WSr);
					/* wake up the Worker process */
				signal_v(WorkerPid, S_EVENT_BNO);
			}
			break;



		case WS_BEGIN_NORM_A & 0xff00:	
#ifdef DEBUG_WS_COMMAND
			printf("Begin Normal Operation (AUTO). \n\r");
#endif
			if (NormalOperation)
			{		/* Set READY bit (PASSED bit should already be set) */
				VXIregs[StatusReg] |= Status_PASSED | Status_READY; 

				WSr.Response = 0x7ffe;
				d_ctrl(DEV_WSsl,MyLA,WSsl_CTRL_Response,&WSr);
			}
			else
			{	WSr.Response = PropagateBNO();
				BNOtype = 2;
				NormalOperation = 1;

					/* Set READY bit (PASSED bit should already be set) */
				VXIregs[StatusReg] |= Status_PASSED | Status_READY; 

				d_ctrl(DEV_WSsl,MyLA,WSsl_CTRL_Response,&WSr);
					/* wake up the Worker process */
				signal_v(WorkerPid, S_EVENT_BNO);
			}
			break;


            case WS_RD_PROT_ERR & 0xff00:
#ifdef DEBUG_WS_COMMAND
                printf("Read Protocol Error ");
#endif
					/* CIs do not need this protocol error handler, handled in driver */
                WSr.Response = PROT_E_NO_ERROR;       
				d_ctrl(DEV_WSsl,MyLA,WSsl_CTRL_Response,&WSr);

#ifdef DEBUG_WS_COMMAND
                printf("0x%x  (ret = 0x%x): ", WSr.Response, WSr.RetVal);
#endif
                break;


		case WS_ANO & 0xff00:
#ifdef DEBUG_WS_COMMAND
			printf("Abort Normal Operation  ");
#endif
			BNOtype = 0;
			NormalOperation = 0;

				/* Handles stoping WS transfers, clears queues,
				 * removes physical devices from your servant list,
				 * enables Asynchronous Mode, disables Response signals. 
				 * NOTE: since no other devices in the system will know 
				 *  about a CI, your Commander's Logical Address is unchanged.
				 */
			WSr.Response = CIAbortNormalOperation(MyLA);

				/* restore default event generation enables */
			CIEventEnables[0] = 0xffffffff;
			CIEventEnables[1] = 0xffffffff;
			CIEventEnables[2] = 0xffffffff;
			CIEventEnables[3] = 0xffffffff;

				/* Empty parser data */
			ClearParser(ParserId);

			/**********
			 *  Note:  To NOT VIOLATE the VXI Specification, you must stop
			 * further action with your servants (now previous servants) and 
			 * not do ANY VXI bus accesses until receiving a Begin Normal 
			 * Operation.  You must also reset any Register Based instruments 
			 * to an known idle state.
			**********/

				/* Clear READY bit (PASSED bit should be kept set) */
			VXIregs[StatusReg] |= Status_PASSED; 

			d_ctrl(DEV_WSsl,MyLA,WSsl_CTRL_Response,&WSr);
#ifdef DEBUG_WS_COMMAND
			printf("0x%x  (ret = 0x%x) \r\n", WSr.Response, WSr.RetVal);
#endif
            break;



		case WS_ENO & 0xff00:
#ifdef DEBUG_WS_COMMAND
			printf("End Normal Operation  ");
#endif
			BNOtype = 0;
			NormalOperation = 0;

				/* handles propagation, waiting for input queues to empty,
				 * and formulating correct response  
				 */
			WSr.Response = CIEndNormalOperation(MyLA);

			/**********
			 *  Note:  To NOT VIOLATE VXI Specification, you must stop
			 * further action with your servants and not do ANY VXI bus
			 * accesses until receiving a Begin Normal Operation.  You must
			 * also set any Register Based instruments to an known idle
			 * state.
			**********/

				/* Clear READY bit (PASSED bit should be kept set) */
			VXIregs[StatusReg] |= Status_PASSED; 

			d_ctrl(DEV_WSsl,MyLA,WSsl_CTRL_Response,&WSr);		/* respond */
#ifdef DEBUG_WS_COMMAND
			printf("0x%x  (ret = 0x%x) \r\n", WSr.Response, WSr.RetVal);
#endif
			break;



		case WS_ID_COMMANDER & 0xff00:
#ifdef DEBUG_WS_COMMAND
			printf("Identify Commander: LA = %d\n\r", data);
#endif
			Cmdr = (int8)data;
			DCIheader.CmdrLogicalAddress = Cmdr;
			break;


		case WS_DEVGRANT & 0xff00:	
#ifdef DEBUG_WS_COMMAND
			printf("Device Grant -- Logical address 0x%x. \n\r", data);
#endif
				/* You have just be assigned a new servant.  Its Logical
				 * address is in the variable 'data' (lower 8 bits of command).
				 * Its Resource Manager entry is valid and you may communicate
				 * with it freely.  Do any necessary signaling to the Worker
				 * to notify it of its existence.
				 */
			break;



		case WS_REL_DEVICE & 0xff00:
#ifdef DEBUG_WS_COMMAND
				printf("Release Device   \r\n");
#endif
				if (DCIheader.CMDRtable[data] == MyLA)
				{		/* This device is my immediate servant.  Handle any local 
						 * freeing of resources here.  (You have no choice whether
						 * to release a device or not).
						 */

				}

					/* Must call CIReleaseDevice to handle special cases and
					 * clean up the Device tables.  It is possible that 
					 * you may receive this command and it not be your servant.
					 * This is handled in the NI function call.
					 */
				WSr.Response =  CIReleaseDevice(MyLA,data);
				d_ctrl(DEV_WSsl,MyLA,WSsl_CTRL_Response,&WSr);
#ifdef DEBUG_WS_COMMAND
				printf("0x%x  (ret = 0x%x) \r\n", WSr.Response, WSr.RetVal);
#endif
			break;



		case WS_READ_PROT & 0xff00:
#ifdef DEBUG_WS_COMMAND
			printf("Read Protocols  ");
#endif
				/* Support WS Trigger, Basic WS, and signal generation */
				/* Note that if full 488.2 functionality is implemented,
				 * PROT_I4 should be added to this list.
				 */
			WSr.Response = 0xffff & ~(PROT_EG | PROT_RG | PROT_TRG | PROT_I);
			d_ctrl(DEV_WSsl,MyLA,WSsl_CTRL_Response,&WSr);
#ifdef DEBUG_WS_COMMAND
			printf("0x%x  (ret = 0x%x) \r\n", WSr.Response, WSr.RetVal);
#endif
			break;


		case WS_RESP_ENA & 0xff00:
				/* keep track of Response interrupts we are enabled for.
				 * The knowledge of this configuration is very important
				 * to the internal workings of the Word Serial driver.
				 */
#ifdef DEBUG_WS_COMMAND
			printf("Response Enable (Control Response): ");
#endif
				/* support DIR and DOR response interrupts */
			respmask = 0x7f & ~(RESP_DIR | RESP_DOR);
				/* Asking for an unsupported bit transition? */
			if ((data & respmask) != respmask)
			{		/* error response but enable supported bits asked for */
				Our_RespEna = (0x7fff & ~(RESP_DIR | RESP_DOR)) | data;
			}
			else
			{   Our_RespEna = (0xffff & ~(RESP_DIR | RESP_DOR)) | data;
			}
			rmtable[MyLA].RespENA = Our_RespEna;
			DCIheader.RespEna = data;
			WSr.Response              = Our_RespEna;
			d_ctrl(DEV_WSsl,MyLA,WSsl_CTRL_Response,&WSr);
#ifdef DEBUG_WS_COMMAND
			printf("0x%x  (ret = 0x%x)\r\n", WSr.Response, WSr.RetVal);
#endif
			break;



		case WS_CTRL_EVENT & 0xff00:
#ifdef DEBUG_WS_COMMAND
			printf("Control Event (%sable 0x%x): ",
				(data & (1<<7))? "Ena":"Dis", data & 0x7f);
#endif
				/* Keep status of enables for EVERY event */
				/* You may change this but make sure your responses to
				 * the command are per VXI Specification
				 */
			eventnum = data & 0x1f;
			eventind = (data >> 5) & 3;

				/* enable or disable? */
			if (data & (1<<7))
				CIEventEnables[eventind] |= (1<<eventnum);
			else
				CIEventEnables[eventind] &= ~(1<<eventnum);

			WSr.Response = 0xfffe;
			d_ctrl(DEV_WSsl,MyLA,WSsl_CTRL_Response,&WSr);
#ifdef DEBUG_WS_COMMAND
			printf("0x%x  (ret = 0x%x)\r\n", WSr.Response, WSr.RetVal);
#endif
			break;


		case WS_ASYN_MODE & 0xff00:
#ifdef DEBUG_WS_COMMAND
			printf("Asynchrous Mode Control (0x%x): ",data);
#endif
				/* Signaler ONLY */
			OurAsynMode = 0xfff0 | (data & 0xf);
			if (!(OurAsynMode & ASYN_RESP_INT))
			{   OurAsynMode |= ASYN_RESP_ENA;         /* 1 = disabled */
				OurAsynMode &= 0x7fff;                /* flag error, not interrupter */
			}
			if (!(OurAsynMode & ASYN_EVENT_INT))
			{   OurAsynMode |= ASYN_EVENT_ENA;        /* 1 = disabled */
				OurAsynMode &= 0x7fff;                /* flag error, not interrupter */
			}
				/* tell commander can send through signals */
			OurAsynMode |= (ASYN_EVENT_INT | ASYN_RESP_INT);

			WSr.Response = OurAsynMode;
			rmtable[MyLA].AsynMode = OurAsynMode;
			DCIheader.AsynMode = OurAsynMode;
			d_ctrl(DEV_WSsl,MyLA,WSsl_CTRL_Response,&WSr);
#ifdef DEBUG_WS_COMMAND
			printf("0x%x  (ret = 0x%x)\r\n", WSr.Response, WSr.RetVal);
#endif
			break;



		case WS_READ_SERV & 0xff00:
#ifdef DEBUG_WS_COMMAND
			printf("Read Servant Area  ");
#endif
				/* This command will never be sent to a CI */
			WSr.Response = OurServantArea;
			d_ctrl(DEV_WSsl,MyLA,WSsl_CTRL_Response,&WSr);

#ifdef DEBUG_WS_COMMAND
			printf("0x%x  (ret = 0x%x): ", WSr.Response, WSr.RetVal);
#endif
			break;


		case WS_BYTERQ & 0xff00:	
		case WS_BYTEAV_0 & 0xff00:	
		case WS_BYTEAV_1 & 0xff00:	
				/* Should NEVER Occur for a DCI */
			break;



			/*******************************************************/
			/*	Handle any other VXI Word Serial Commands HERE.
			 * Any other commands not implemented will automatically
			 * fall thru to the default switch which will send 
			 * the Unrecognized Command Event for you.
			 *
			 * NOTE: User Defined commands are split into National
			 * Instruments Reserved commands and DCI user commands
			 * according to bit 14 (1 = NI, 0 = DCI).  Do NOT use
  			 * any National Instruments reserved Word Serial commands
			 * unless expressly told to do so.
			 */
			/*******************************************************/



		default:	
#ifdef DEBUG_WS_COMMAND
			if (com & 0x8000)
			{ 	
				printf("VXIbus reserved command --> 0x%x\n\r", com);
			}
			else
			{ 	
				printf("User defined command --> 0x%x\n\r", com);
			}
#endif
				/* Send Back UnRecognized Command */
			SignalOut(Cmdr, (SIG_UNREC_CMD | MyLA));

			break;
	}
}	



/*
 *	The following function is used to propagate Begin Normal Operation to all
 * of our message based servants.
 */
PropagateBNO()
{
	int8 *la, isone;
    uint16 IdCom, *RegBase;
	WScom WSc;
	uint32 BNOreturn;


	BNOreturn = 0xfffe;

	if (!DCIheader.PropagateBNO)
	{	DCIheader.PropagateBNO = 1;
		return BNOreturn;
	}

#ifdef DCI_BNO_DEBUG
	printf("Sending Begin Normal Operation to DCI %d Immediate Message Based servants\n\r",MyLA);
#endif

    isone = 0;
		/* Must tell servant that commander is GPIB-VXI NOT!!! the DCI */
    IdCom = (uint16)(WS_ID_COMMANDER | DCIheader.GPIB_VXI_LA);
	la = &DCIheader.ServantList[rmtable[MyLA].servants];
    for (; *la != 0xff; la++)
    { 	RegBase = VXIregBase(*la);
		if (((RegBase[IdReg] & DEV_CLASS_MASK) >> 14) == DEV_CLASS_MBD)
        {
#ifdef DCI_BNO_DEBUG
            printf(" %d",*la);
#endif

				/* Send Identify Commander but only to Bus Masters */
			if (RegBase[ProtocolReg] & Protocol_MASTER)
			{	WSc.Command = (uint16)IdCom;
				WSc.mode = 0;
				d_ctrl(DEV_WS,*la,WS_CTRL_Command,&WSc);
				if (!(WSc.RetVal & WS_RET_IODONE))
    			{ 	printf("DCI %d ERROR: LA %d errored when sent Identify Commander.\r\n",
						MyLA, *la);
				}
			}

			if (rmtable[*la].prot & PROT_SPEC_1_2)
			{ 		/* Send Begin Normal Operation command */
				WSc.Command = (uint16)WS_BEGIN_NORM;
				WSc.mode = 0;
				d_ctrl(DEV_WS,*la,WS_CTRL_Command,&WSc);
				if (!(WSc.RetVal & WS_RET_IODONE))
				{	BNOreturn = 0x5b00 | *la;				/* device failed to accept */
				}
			}
			else
			{ 		/* Send Begin Normal Operation query */
				WSc.Command = (uint16)WS_BEGIN_NORM;
				WSc.mode = WS_MODE_RESPONSE;
				d_ctrl(DEV_WS,*la,WS_CTRL_Command,&WSc);
				if (!(WSc.RetVal & WS_RET_IODONE))
				{	BNOreturn = 0x5b00 | *la;				/* device failed to accept */
				}
				else
                {   if ((WSc.Response & 0xf000) != 0xf000)	/* device gave error response */
					{	BNOreturn = WSc.Response & 0xf0ff;
						if ((WSc.Response & 0x0f00) == 0x0f00)
							BNOreturn |= 0x0f00;
						else
							BNOreturn |= 0x0b00;
					}
				}
			}

            isone = 1;
#ifdef DCI_BNO_DEBUG
            printf("!");
#endif
        }
    }

#ifdef DCI_BNO_DEBUG
	if (isone)
   	 	printf(" Done.\n\r");
   	else
		printf("  No Immediate DCI %d Message Based servants.\n\r",MyLA);
#endif

	return BNOreturn;
}



/*
 *   This function is used to abort Word Serial Operations with Message 
 * servants as well as send Word Serial Clear command to them.  Sending
 * of the Word Serial Clear command after the Word Serial Abort is NOT
 * required by the VXI Specification and this code may be modified to 
 * reflect that if desired.
 */
ClearServants()
{
	int8 *la, isone;
    uint16 *RegBase;
	WScom WSc;


#ifdef DCI_CLEAR_DEBUG
	printf("DCI %d clearing servants: \r\n", MyLA);
#endif

    isone = 0;
	la = &DCIheader.ServantList[rmtable[MyLA].servants];
    for (; *la != 0xff; la++)
    { 	RegBase = VXIregBase(*la);
		if (((RegBase[IdReg] & DEV_CLASS_MASK) >> 14) == DEV_CLASS_MBD)
        {
#ifdef DCI_CLEAR_DEBUG
            printf(" %d",*la);
#endif

				/* Device Clear Abort Any WS I/O transfer with Servant */
			d_ctrl(DEV_WS,*la,WS_CTRL_Clear,0);

				/* Send Word Serial Device Clear to Servant */
			WSc.Command = (uint16)WS_CLEAR;
			WSc.mode = 0;
			d_ctrl(DEV_WS,*la,WS_CTRL_Command,&WSc);
			if (!(WSc.RetVal & WS_RET_IODONE))
    		{ 	printf("DCI %d ERROR: Sending Device Clear.\r\n",MyLA);
				return;
			}

            isone = 1;
#ifdef DCI_CLEAR_DEBUG
            printf("!");
#endif
        }
    }

#ifdef DCI_CLEAR_DEBUG
	if (isone)
   	 	printf(" Done.\n\r");
   	else
		printf("  No Immediate DCI %d Message Based servants.\n\r",MyLA);
#endif
}



/*
 *   The following is used get rid of any pending messages on the Asynch 
 * exchange that have not yet been processed.  It simply pulls off messages 
 * continuously until there are no more to be pulled off.  This should be
 * used in situations in which the CI is in an invalid situation and needs
 * to reset its entire state.  This should NOT be done in the Word Serial
 * Clear handler.  When the clear command is received as previously sent 
 * commands have be received and processed (the exchange is a fifo) so any
 * more messages on the exchange are new commands sent AFTER the clear.
 */
ClearExchange()
{
	struct { uint32 u,h, type, p1, p2, p3} m;		/* pSOS message struct */

	while (req_x(AsynchXid,&m,0,2) == 0)
		;		

}


